from urllib.parse import ParseResult
import yaml
import logging
import subprocess

from pathlib import Path
from tuxtrigger.inputvalidation import YamlValidator


LOG = logging.getLogger("tuxtrigger")


def yaml_load(data):
    return yaml.load(data, Loader=yaml.FullLoader)


def yaml_file_read(file_path: Path):
    with file_path.open("r") as file:
        loaded_yaml = yaml_load(file)
    return loaded_yaml


def yaml_file_write(data_input, file_path: Path):
    with file_path.open("w") as writer:
        yaml.dump(data_input, writer)


def create_repo_list(data_input):
    if data_input is None:
        return 0
    for item in data_input["repositories"]:
        yield YamlValidator(**item)


# def find_in_yaml(value, data_yaml):
#     if data_yaml is None:
#         return None
#     if value in data_yaml:
#         return data_yaml[value]
#     for k, v in data_yaml.items():
#         if isinstance(v, dict):
#             for i in find_in_yaml(value, v):
#                 return i


def compare_sha(url_key, branch_name, input_dict, data_yaml_file) -> bool:
    if data_yaml_file is None:
        LOG.warning("\t*** Data Input is none, not able to compare sha")
        LOG.debug(f'\t-> sha: {input_dict[branch_name]["sha"]}')
        return False
    if branch_name not in data_yaml_file[url_key]:
        LOG.warning("\t*** Branch not found in yaml file")
        LOG.debug(f"\t*** branch name: {branch_name}")
        return False
    if (
        not input_dict[branch_name]["sha"]
        == data_yaml_file[url_key][branch_name]["sha"]
    ):
        LOG.info(
            f'\t-> sha: {input_dict[branch_name]["sha"]} vs \
        previous sha {data_yaml_file[url_key][branch_name]["sha"]}'
        )
        return True
    LOG.info(f'\t-> sha: {input_dict[branch_name]["sha"]}')
    LOG.info("\t-> no changes")
    return False


def create_squad_project(url: ParseResult, branch_name: str) -> str:
    url_fragment = (url.path).split("/")
    url_fragment = list(filter(None, url_fragment))
    return f'{(url_fragment[-1]).replace(".git","")}-{branch_name.replace("/","-")}'


def run_lsremote(repository: str, branch: str) -> dict:
    value_dict = dict()
    git_result = subprocess.run(
        ["git", "ls-remote", repository, branch],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    splited_value = git_result.stdout.split()
    value_dict[branch] = {"sha": "", "ref": ""}
    if git_result.returncode != 0:
        return value_dict

    if splited_value:
        value_dict[branch] = {"sha": splited_value[0], "ref": splited_value[1]}

    return value_dict



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import os
import subprocess
import json
import logging
import tempfile
import requests


LOG = logging.getLogger("tuxtrigger")


def tux_console_build(git_repo, git_ref, target_arch, kconfig, toolchain):
    with tempfile.NamedTemporaryFile(suffix=".json") as json_temp:
        build = subprocess.run(
            [
                "tuxsuite",
                "build",
                "--git-repo",
                git_repo,
                "--git-ref",
                git_ref,
                "--target-arch",
                target_arch,
                "--kconfig",
                kconfig,
                "--toolchain",
                toolchain,
                "--json-out",
                json_temp.name,
                "--no-wait",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        if build.returncode != 0:
            LOG.warning(f"*build stdout {build.stdout}")
            LOG.warning(f"*build stderr {build.stderr}")
            raise Exception(f"*** Tuxsuite not build repo {build.stderr}")

        LOG.debug(build.stdout)

        json_output = json.load(json_temp)
        LOG.debug(f'\t*** Build UID: {json_output["uid"]}')
        return json_output["uid"]


def tux_console_plan(json_data, plan_file, squad_group, squad_project) -> int:
    if json_data is None:
        LOG.warning("\t** Not able to submit plan -> json output is None")
        return 1
    with tempfile.NamedTemporaryFile(suffix=".json") as json_temp:
        plan = subprocess.run(
            [
                "tuxsuite",
                "plan",
                "--git-repo",
                json_data["git_repo"],
                "--git-ref",
                json_data["git_ref"],
                "--name",
                json_data["git_describe"],
                "--no-wait",
                plan_file,
                "--json-out",
                json_temp.name,
            ]
        )
        if plan.returncode != 0:
            LOG.warning(f'\t** Submiting Plan for {json_data["git_describe"]} failed')
            return 1
        LOG.info(f'\t-> Submiting Plan for {json_data["git_describe"]}')
        squad_submit(json_data, squad_group, squad_project, json_temp.name)
        return 0


def build_result(uid):
    if uid is None:
        return None
    build = subprocess.run(
        ["tuxsuite", "build", "get", uid, "--json"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    json_output = json.loads(build.stdout)
    LOG.debug(f"\t ** JSON OUTPUT: {json_output}")
    LOG.info(
        f'\t-> Build {json_output["uid"]} state: {json_output["state"]}, result: {json_output["result"]}, git describe: {json_output["git_describe"]}'
    )
    return json_output


def build_result_list() -> int:
    build = subprocess.run(
        ["tuxsuite", "build", "list", "--json"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    json_output_list = json.loads(build.stdout)
    for item in json_output_list:
        LOG.info(
            f'\t-> Build {item["uid"]} state: {item["state"]}, result: {item["result"]}, git describe: {item["git_describe"]}'
        )
    return 0


def squad_submit(json_data, squad_group, squad_project, plan_json):
    if squad_group is None or squad_project is None:
        LOG.warn("** SQUAD config is not available! Unable to process **")
        return 1

    squad = subprocess.run(
        [
            "squad-client",
            "submit-tuxsuite",
            "--group",
            squad_group,
            "--project",
            squad_project,
            "--build",
            json_data["git_describe"],
            "--backend",
            "tuxsuite.com",
            "--json",
            plan_json,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    if squad.returncode != 0:
        LOG.warning(f"*build stdout {squad.stdout}")
        LOG.warning(f"*build stderr {squad.stderr}")
        raise Exception(f"*** squad-client not able to pass data {squad.stderr}")

    LOG.debug(squad.stdout)
    LOG.info(f'\t-> Plan submitted to SQUAD git describe: {json_data["git_describe"]}')
    return 0


def squad_sha_request(squad_project):
    url = f'{os.getenv("SQUAD_HOST")}/api/projects/?slug={squad_project}'
    resp = requests.get(url)
    json_output = json.loads(resp.content)
    id = json_output["results"][0]["id"]
    build_url = (
        f'{os.getenv("SQUAD_HOST")}/api/projects/{id}/builds/?ordering=-datetime'
    )
    build_resp = requests.get(build_url)
    build_json = json.loads(build_resp.content)
    lastest_version_url = build_json["results"][0]["metadata"]
    metadata_resp = requests.get(lastest_version_url)
    metadata_json = json.loads(metadata_resp.content)
    git_sha = metadata_json["git_sha"]
    return git_sha



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT


class TuxTriggerException(Exception):
    pass
    # def __str__(self):
    #     name = super().__str__()
    #     if hasattr(self, "msg"):
    #         return self.msg.format(name=name)
    #     else:
    #         return name


class InvalidArgument(TuxTriggerException):
    pass



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Dict, List

from urllib.parse import urlparse


class Base:
    def url_validation(self, url):
        urlScheme = urlparse(url)
        if urlScheme.scheme not in ["http", "https"]:
            raise TypeError("Invalid url input.")

    def url_path(self):
        return urlparse(self.url)


@dataclass
class YamlValidator(Base):

    url: str
    squad_group: str
    branches: List[Dict]

    def __post_init__(self):
        self.url_validation(self.url)



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import contextlib
import logging
import sys
import logging.handlers
import time

from tuxtrigger.argparser import setup_parser
from tuxtrigger.builder import build_result, tux_console_build, tux_console_plan
from tuxtrigger.yamlload import (
    yaml_file_read,
    create_repo_list,
    yaml_file_write,
    compare_sha,
    create_squad_project,
    run_lsremote,
)

LOG = logging.getLogger("tuxtrigger")

build_params = {
    "git_repo": "",
    "git_ref": "",
    "target_arch": "x86_64",
    "toolchain": "gcc-12",
    "kconfig": "tinyconfig",
}


def main() -> int:
    parser = setup_parser()
    options = parser.parse_args()
    output_file = (options.output).absolute()
    output_file.parent.mkdir(exist_ok=True)
    plan_path = (options.plan).absolute()

    log_handler = logging.StreamHandler(sys.stdout)
    log_handler.setFormatter(logging.Formatter("%(message)s"))
    log_file_handler = logging.handlers.WatchedFileHandler(
        (options.log_file).absolute(),
        mode="w",
    )
    LOG.addHandler(log_handler)
    LOG.addHandler(log_file_handler)

    if options.log_level == "debug":
        LOG.setLevel(logging.DEBUG)
    elif options.log_level == "warn":
        LOG.setLevel(logging.WARNING)
    elif options.log_level == "error":
        LOG.setLevel(logging.ERROR)
    else:
        LOG.setLevel(logging.INFO)

    if options.config:
        uid_queue = list()
        output_data = dict()
        squad_config = dict()
        config_yaml_data = yaml_file_read(options.config)
        repo_list = create_repo_list(config_yaml_data)
        archive_data_file = None
        with contextlib.suppress(FileNotFoundError):
            archive_data_file = yaml_file_read(output_file)
        LOG.info("Checking repositories:")
        for repo in repo_list:
            value_dict = dict()
            repo_name = repo.url_path()
            group_name = repo.squad_group
            LOG.info(f"* {repo.url}")
            for branch in repo.branches:
                LOG.info(f' * branch: {branch["name"]}')
                build_params["git_repo"] = repo.url
                build_params["git_ref"] = branch["name"]
                ls_remote_result = run_lsremote(repo.url, branch["name"])

                repo_changed = compare_sha(
                    repo_name.geturl(),
                    branch["name"],
                    ls_remote_result,
                    archive_data_file,
                )
                value_dict.update(ls_remote_result)
                if options.submit == "never":
                    LOG.info("** Builds suspended **")

                elif repo_changed or options.submit == "always":
                    build_uid = tux_console_build(**build_params)
                    uid_queue.append(build_uid)
                    squad_config[build_uid] = (
                        branch["plan"],
                        group_name,
                        branch.get(
                            "squad_project",
                            create_squad_project(repo.url_path(), branch["name"]),
                        ),
                    )
            output_data[repo_name.geturl()] = value_dict

        LOG.info("** Build Phase Completed **")

        while uid_queue:
            for uid in uid_queue:
                time.sleep(10)
                json_build_output = build_result(uid)
                if json_build_output["git_describe"] is not None:
                    LOG.debug(f"\t*UID: {uid}")
                    LOG.debug(f"\t*Plan file name: {squad_config[uid][0]}")
                    LOG.debug(f"\t*SQUAD group name: {squad_config[uid][1]}")
                    LOG.debug(f"\t*SQUAD project name: {squad_config[uid][2]}")
                    tux_console_plan(
                        json_build_output,
                        plan_path / squad_config[uid][0],
                        squad_config[uid][1],
                        squad_config[uid][2],
                    )
                    uid_queue.remove(uid)
        LOG.info("** Submiting Plans Phase Completed **")

        yaml_file_write(output_data, output_file)
        LOG.info("** SHA List Updated **")
    else:
        LOG.info("* Please input path to config yaml file!")
        parser.print_usage()

    return 0


def start() -> None:
    if __name__ == "__main__":
        sys.exit(main())


start()



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import sys
import pytest

from pathlib import Path
import tuxtrigger.__main__ as main_module

BASE = (Path(__file__) / "../..").resolve()

VALUE_DICT = {
    "master": {"sha": "12345", "ref": "refs/master/torvalds"},
    "linux-5.18.y": {"sha": "12345", "ref": "refs/v5.19"},
    "queue/5.4": {"sha": "12345", "ref": "refs/for-laurent"},
}

UID = "2CCI3BkwKdqM4wOOwB5xRRxvOha"

JSON_OUTPUT = {
    "empty": "no_real_values",
    "uid": "1234",
    "git_repo": "https://gitlab.com/no_real_repo",
    "git_ref": "master",
    "git_describe": "v5.19-rc7",
}


@pytest.fixture
def argv():
    return ["tuxtrigger"]


@pytest.fixture(autouse=True)
def patch_argv(monkeypatch, argv):
    monkeypatch.setattr(sys, "argv", argv)


class TestMain:
    def test_start(self, monkeypatch, mocker):
        monkeypatch.setattr(main_module, "__name__", "__main__")
        main = mocker.patch("tuxtrigger.__main__.main", return_value=1)
        exit = mocker.patch("sys.exit")
        main_module.start()
        main.assert_called()
        exit.assert_called_with(1)

    def test_main_version(self, argv, capsys):
        argv.append("-v")
        with pytest.raises(SystemExit):
            main_module.main()
        out, out_err = capsys.readouterr()
        assert "TuxTrigger" in out

    def test_main_correct_data(self, argv, monkeypatch, mocker):
        argv.extend([f"{BASE}/test/test_files/happy_path.yaml"])
        monkeypatch.setattr(main_module, "__name__", "__main__")
        main_lsremote = mocker.patch(
            "tuxtrigger.__main__.run_lsremote", return_value=VALUE_DICT
        )
        main_build_uid = mocker.patch(
            "tuxtrigger.__main__.tux_console_build", return_value=UID
        )
        main_build_output = mocker.patch(
            "tuxtrigger.__main__.build_result", return_value=JSON_OUTPUT
        )
        main_compare_sha = mocker.patch(
            "tuxtrigger.__main__.compare_sha", return_value=True
        )
        exit = mocker.patch("sys.exit")
        main_module.start()
        main_lsremote.assert_called()
        main_build_uid.assert_called()
        main_compare_sha.assert_called()
        main_build_output.assert_called()
        exit.assert_called_with(0)

    def test_main_key_error(self, argv):
        argv.extend([f"{BASE}/test/test_files/error_path.yaml"])
        with pytest.raises(KeyError):
            main_module.main()
            assert main_module.main() == 1

    def test_main_build_never(self, argv, capsys, monkeypatch, mocker):
        argv.append(f"{BASE}/test/test_files/happy_path.yaml")
        monkeypatch.setattr(main_module, "__name__", "__main__")
        main = mocker.patch("tuxtrigger.__main__.run_lsremote", return_value=VALUE_DICT)
        argv.append("--submit=never")
        main_module.main()
        out, out_err = capsys.readouterr()
        main.assert_called()
        assert "** Builds suspended **" in out

    def test_main_build_always(self, argv, capsys, mocker, monkeypatch):
        argv.append(f"{BASE}/test/test_files/happy_path.yaml")
        argv.append("--submit=always")
        monkeypatch.setattr(main_module, "__name__", "__main__")
        main_lsremote = mocker.patch(
            "tuxtrigger.__main__.run_lsremote", return_value=VALUE_DICT
        )
        main_build_uid = mocker.patch(
            "tuxtrigger.__main__.tux_console_build", return_value=UID
        )
        main_build_output = mocker.patch(
            "tuxtrigger.__main__.build_result", return_value=JSON_OUTPUT
        )
        main_module.main()
        main_lsremote.assert_called()
        main_build_uid.assert_called()
        main_build_output.assert_called()
        out, out_err = capsys.readouterr()
        assert "** Submiting Plan for" in out



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT


import pytest
from unittest.mock import MagicMock
from tuxtrigger import yamlload

from tuxtrigger.inputvalidation import YamlValidator
from tuxtrigger.yamlload import (
    yaml_file_read,
    yaml_file_write,
    compare_sha,
)
from pathlib import Path

BASE = (Path(__file__) / "..").resolve()

ERROR_PATH = BASE / "./test_files/error_path.yaml"
HAPPY_PATH = BASE / "./test_files/happy_path.yaml"
FILE_TO_CREATE = BASE / "./test_files/new_file.yaml"

VALUE_DICT = {
    "linux-5.18.y": {
        "sha": "2437f53721bcd154d50224acee23e7dbb8d8c622",
        "ref": "refs/heads/linux-5.18.y",
    }
}
VALUE = "linux-5.18.y"
RIGHT_KEY = "https://gitlab.com/Linaro/lkft/mirrors/stable/linux-stable"
WRONG_KEY = "/linux-not-existing"


def test_yaml_file_read():
    assert type(yaml_file_read(ERROR_PATH)) == dict


def test_create_repo_list(repo_setup_good):
    assert isinstance(next(repo_setup_good), YamlValidator) is True
    assert repo_setup_good is not None


def test_yaml_file_write(tmpdir):
    test_input = yaml_file_read(ERROR_PATH)
    yaml_file_write(test_input, (tmpdir / "test.yaml"))
    assert (tmpdir / "test.yaml").exists()


# def test_find_in_yaml(correct_archive_read, wrong_archive_read):
#     assert {
#         "ref": "refs/heads/linux-5.18.y",
#         "sha": "2437f53721bcd154d50224acee23e7dbb8d8c62b",
#     } == find_in_yaml(VALUE, correct_archive_read)
#     assert find_in_yaml(VALUE, wrong_archive_read) is None


def test_compare_sha_correct(correct_archive_read):
    with pytest.raises(KeyError):
        compare_sha(WRONG_KEY, VALUE, VALUE_DICT, correct_archive_read)
    assert compare_sha(RIGHT_KEY, VALUE, VALUE_DICT, correct_archive_read) is True


def test_compare_sha_wrong(wrong_archive_read):
    assert compare_sha(WRONG_KEY, VALUE, VALUE_DICT, wrong_archive_read) is False
    assert compare_sha(RIGHT_KEY, VALUE, VALUE_DICT, wrong_archive_read) is False


def test_run_lsremote():
    yamlload.run_lsremote = MagicMock(
        return_value={"master": {"ref": "123", "sha": "456"}}
    )
    assert yamlload.run_lsremote(
        "https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git", "master"
    ) == {"master": {"ref": "123", "sha": "456"}}



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT

from unittest.mock import MagicMock
import pytest

from pathlib import Path
from tuxtrigger import builder
from tuxtrigger.builder import (
    squad_submit,
    tux_console_build,
    tux_console_plan,
    build_result,
)

PARAMS = {
    "git_repo": "https://gitlab.com/Linaro/lkft/mirrors/stable/linux-stable-rc",
    "git_ref": "master",
    "target_arch": "x86_64",
    "toolchain": "gcc-12",
    "kconfig": "tinyconfig",
}

FAIL_PARAMS = {
    "git_repo": "https://gitlab.com/not_existing_repo",
    "git_ref": "not_existing_branch",
    "target_arch": "x86_64",
    "toolchain": "gcc-12",
    "kconfig": "tinyconfig",
}

BASE = (Path(__file__) / "..").resolve()

PLAN = BASE / "test_files/planTest.yaml"
PLAN_FAIL = BASE / "test_files/planTestc.yaml"
JSON_PLAN_RESULT = BASE / "test_files/plan_result.json"

UID = "2CCI3BkwKdqM4wOOwB5xRRxvOha"


def test_tux_console_build():
    uid_value = tux_console_build(**PARAMS)
    assert uid_value is not None
    assert len(uid_value) == 27


def test_tux_console_build_mock():
    builder.tux_console_build = MagicMock(return_value=UID)
    assert builder.tux_console_build() == UID


def test_tux_console_plan(squad_group_good, squad_project_good):
    json_data = None
    assert tux_console_plan(json_data, PLAN, squad_group_good, squad_project_good) == 1


def test_build_result():
    json_output = build_result(UID)
    assert build_result(None) is None
    assert isinstance(json_output, dict)


def test_build_result_list():
    builder.build_result_list = MagicMock(return_value=0)
    assert builder.build_result_list() == 0


def test_tux_console_build_error(capsys):
    with pytest.raises(Exception):
        fail_build = tux_console_build(**FAIL_PARAMS)
        out, out_err = capsys.readouterr()
        assert fail_build != 0
        assert "Tuxsuite not build" in out


def test_tux_console_plan_error(squad_group_good, squad_project_good):
    json_data = {
        "empty": "no_real_values",
        "uid": "1234",
        "git_repo": "https://gitlab.com/no_real_repo",
        "git_ref": "master",
        "git_describe": "v5.19-rc7",
    }
    fail_plan = tux_console_plan(
        json_data, PLAN_FAIL, squad_group_good, squad_project_good
    )
    assert fail_plan != 0


def test_squad_submit(squad_env_setup, squad_group_good, squad_project_good):
    json_data = {
        "empty": "no_real_values",
        "uid": "1234",
        "git_repo": "https://gitlab.com/no_real_repo",
        "git_ref": "master",
        "git_describe": "v5.19-rc7",
    }
    with pytest.raises(Exception):
        squad_submit(json_data, squad_group_good, squad_project_good, JSON_PLAN_RESULT)



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import pytest

from pathlib import Path
from tuxtrigger.yamlload import yaml_file_read, create_repo_list

BASE = (Path(__file__) / "..").resolve()

ERROR_PATH = BASE / "./test_files/error_path.yaml"
HAPPY_PATH = BASE / "./test_files/happy_path.yaml"
GITSHA_FILE = BASE / "./test_files/gitsha.yaml"


@pytest.fixture
def wrong_archive_read():
    try:
        return yaml_file_read(Path("non_existing_file"))
    except FileNotFoundError:
        return None


@pytest.fixture
def correct_archive_read():
    return yaml_file_read(GITSHA_FILE)


@pytest.fixture
def repo_setup_error():
    return create_repo_list(yaml_file_read(ERROR_PATH))


@pytest.fixture
def repo_setup_good():
    return create_repo_list(yaml_file_read(HAPPY_PATH))


@pytest.fixture
def squad_group_good():
    return "~non.existing"


@pytest.fixture
def squad_project_good():
    return "example_project"


@pytest.fixture(autouse=True)
def squad_env_setup(monkeypatch):
    monkeypatch.setenv("SQUAD_TOKEN", "some-value")



#
# Copyright 2022-present Linaro Limited
#
# SPDX-License-Identifier: MIT

from tuxtrigger.exceptions import InvalidArgument, TuxTriggerException


def test_tux_trig_exception():
    exc = TuxTriggerException()
    assert isinstance(exc, Exception) is True


def test_invalid_args():
    arg_exc = InvalidArgument()
    assert isinstance(arg_exc, TuxTriggerException) is True
